import xml.etree.ElementTree as ET
from contextlib import closing

from pyjabber.db.database import connection
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Roster(metaclass=Singleton):
    def __init__(self, db_connection_factory=None) -> None:
        self._handlers = {
            "get": self.handle_get,
            "set": self.handle_set,
            "result": self.handle_result
        }
        self._ns = {
            "ns": "jabber:iq:roster",
            "query": "{jabber:iq:roster}query",
            "item": "{jabber:iq:roster}item"
        }
        self._db_connection_factory = db_connection_factory or connection

    def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        return self._handlers[element.attrib["type"]](jid, element)

    ############################################################
    ############################################################

    def handle_get(self, jid: JID, element: ET.Element):
        jid = jid.bare()

        with closing(self._db_connection_factory()) as con:
            res = con.execute("SELECT * FROM roster WHERE jid = ?", (jid,))
            roster = res.fetchall()

        iq_res = IQ(type=IQ.TYPE.RESULT.value, id=element.attrib["id"])
        query = ET.SubElement(iq_res, "query", attrib={"xmlns": self._ns["ns"]})

        for item in roster:
            query.append(ET.fromstring(item[-1]))

        return ET.tostring(iq_res)

    def handle_set(self, jid: JID, element: ET.Element):
        query = element.find(self._ns["query"])
        jid = jid.bare()

        if query is None:
            return SE.invalid_xml()

        new_item = query.findall(self._ns["item"])

        if len(new_item) != 1:
            return SE.invalid_xml()

        new_item = new_item[0]
        remove = None
        if "subscription" in new_item.attrib.keys():
            remove = new_item.attrib["subscription"] == "remove"

        with closing(self._db_connection_factory()) as con:
            res = con.execute("SELECT * FROM roster WHERE jid = ?", (jid,))
            res = res.fetchall()

            roster = res
            roster = [ET.fromstring(r[-1]) for r in roster]
            match_item = [i for i in roster if i.attrib["jid"] == new_item.attrib["jid"]]

            if match_item:
                # Delete roster item
                if remove:
                    con.execute("""
                                DELETE FROM roster
                                WHERE jid = ? AND rosterItem = ?
                                """,
                                (jid,
                                 ET.tostring(match_item[0]).decode()))
                    con.commit()

                else:
                    con.execute("""
                                UPDATE roster
                                SET rosterItem = ?
                                WHERE jid = ? AND rosterItem = ?
                                """,
                                (ET.tostring(new_item).decode(),
                                 jid,
                                 ET.tostring(match_item[0])))
                    con.commit()

            else:
                if not remove:
                    con.execute(
                        "INSERT INTO roster(jid, rosterItem) VALUES (?, ?)",
                        (jid,
                         ET.tostring(new_item).decode()))
                    con.commit()

            res = ET.Element(
                "iq",
                attrib={
                    "id": element.attrib["id"],
                    "type": "result"
                }
            )

            return ET.tostring(res)

    def handle_result(self, element: ET.Element):
        # It's safe to ignore this stanza
        return



import asyncio
import os
import ssl

from loguru import logger
from xml import sax

from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.XMLParser import XMLParser

FILE_AUTH = os.path.dirname(os.path.abspath(__file__))


class XMLProtocol(asyncio.Protocol):
    """
    Protocol to manage the network connection between nodes in the XMPP network. Handles the transport layer.

    :param namespace: namespace of the XML tags (jabber:client or jabber:server)
    :param host: Host for connections
    :param connection_timeout: Max time without any response from a client. After that, the server will terminate the connection
    :param cert_path: Path to custom domain certs. By default, the server generates its own certificates for hostname
    :param enable_tls1_3: Boolean. Enables the use of TLSv1.3 in the STARTTLS process
    """

    def __init__(
            self,
            namespace,
            host,
            connection_timeout,
            cert_path,
            enable_tls1_3=False):

        self._xmlns = namespace
        self._host = host
        self._connection_timeout = connection_timeout
        self._connection_manager = ConnectionManager()
        self._cert_path = cert_path
        self._enable_tls1_3 = enable_tls1_3

        self._transport = None
        self._peer = None
        self._xml_parser = None
        self._timeout_monitor = None

    def connection_made(self, transport):
        """
        Called when a client or another server opens a TCP connection to the server

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        if transport:
            self._transport = transport
            self._peer = self._transport.get_extra_info('peername')

            self._xml_parser = sax.make_parser()
            self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
            self._xml_parser.setFeature(sax.handler.feature_external_ges, False)
            self._xml_parser.setContentHandler(
                XMLParser(
                    self._host,
                    self._transport,
                    self.task_tls,
                )
            )

            if self._connection_timeout:
                self._timeout_monitor = StreamAlivenessMonitor(
                    timeout=self._connection_timeout,
                    callback=self.connection_timeout
                )

            self._connection_manager.connection(self._peer, self._transport)

            logger.info(f"Connection from {self._peer}")
        else:
            logger.error("Invalid transport")

    def connection_lost(self, exc):
        """
        Called when a client or another server closes a TCP connection to the server

        :param exc: Exception that caused the connection to close
        :type exc: Exception
        """
        if self._transport:
            logger.info(f"Connection lost from {self._peer}: Reason {exc}")

            self._transport = None
            self._xml_parser = None

    def data_received(self, data):
        """
        Called when data is received from the client or another server

        :param data: Chunk of data received
        :type data: Byte array
        """
        try:
            logger.debug(f"Data received from <{self._peer}>: {data.decode()}")
        except:
            logger.debug(f"Binary data recived")

        if self._timeout_monitor:
            self._timeout_monitor.reset()

        '''
        Some XMPP clients/libraries send the XML header
        with the stream stanza all together.
        This can lead sometimes in a exception for the parser.
        An XML header gives no usefull information, as the RFC6120
        takes the vesion 1.0 as standar, so we can remove it safely.
        If I try to remove the header inside the except block, the parser
        keeps refusing it. I'm forced to do always before the feed.
        I probably should change the parser
        '''
        data = data.replace(b"<?xml version=\'1.0\'?>", b"")
        data = data.replace(b"<?xml version=\"1.0\"?>", b"")

        self._xml_parser.feed(data)

    def eof_received(self):
        """
        Called when the client or another server sends an EOF
        """
        if self._transport:
            logger.debug(f"EOF received from {self._peer}")
            self._connection_manager.disconnection(self._peer)

    def connection_timeout(self):
        """
        Called when the stream is not responding for a long tikem
        """
        logger.debug(f"Connection timeout from {self._peer}")

        self._transport.write("<connection-timeout/>".encode())
        self._transport.close()

        self._connection_manager.disconnection(self._peer)

        self._transport = None
        self._xml_parser = None

    ###########################################################################
    ###########################################################################
    ###########################################################################

    def task_tls(self):
        """
            Sync function to call the STARTTLS coroutine
        """
        loop = asyncio.get_running_loop()
        asyncio.ensure_future(self.enable_tls(loop), loop=loop)

    async def enable_tls(self, loop):
        """
            Coroutine to upgrade the connection to TLS
            It swaps the transport for the XMLProtocol, and XMLParser

            :param loop: Running asyncio loop
        """
        parser = self._xml_parser.getContentHandler()

        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        if not self._enable_tls1_3:
            ssl_context.options |= ssl.OP_NO_TLSv1_3

        if self._cert_path:
            ssl_context.load_cert_chain(
                certfile=os.path.join(self._cert_path, f"{self._host}_cert.pem"),
                keyfile=os.path.join(self._cert_path, "certs", f"{self._host}_key.pem"),
            )
        else:
            ssl_context.load_cert_chain(
                certfile=os.path.join(FILE_AUTH, "certs", f"{self._host}_cert.pem"),
                keyfile=os.path.join(FILE_AUTH, "certs", f"{self._host}_key.pem"),
            )

        try:
            new_transport = await loop.start_tls(
                transport=self._transport,
                protocol=self,
                sslcontext=ssl_context,
                server_side=True)

            self._transport = new_transport
            parser.buffer = self._transport
            logger.debug(f"Done TLS for <{self._peer}>")

        except ConnectionResetError:
            logger.error(f"ERROR DURING TLS UPGRADE WITH <{self._peer}>")
            if not self._transport.is_closing():
                self._connection_manager.close(self._peer)



import pytest
from unittest.mock import MagicMock, patch
import asyncio
import ssl
from pyjabber.network.server.outcoming.XMLServerOutcomingProtocol import XMLServerOutcomingProtocol
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor

@pytest.fixture
def setup_protocol():
    namespace = 'namespace'
    host = 'example.com'
    public_host = 'public.example.com'
    connection_timeout = 10
    connection_manager = MagicMock()
    queue_message = MagicMock()
    protocol = XMLServerOutcomingProtocol(
        namespace,
        host,
        public_host,
        connection_timeout,
        connection_manager,
        queue_message
    )
    return protocol, connection_manager

def test_initialization(setup_protocol):
    protocol, connection_manager = setup_protocol
    assert protocol._host == 'example.com'
    assert protocol._public_host == 'public.example.com'
    assert protocol._connection_timeout == 10
    assert protocol._connection_manager == connection_manager
    assert protocol._queue_message is not None

@patch('xml.sax.make_parser', return_value=MagicMock())
def test_connection_made_with_transport(mock_make_parser, setup_protocol):
    protocol, connection_manager = setup_protocol
    transport = MagicMock()
    transport.get_extra_info.return_value = 'peername'

    protocol.connection_made(transport)

    assert protocol._transport == transport
    mock_make_parser.return_value.setContentHandler.assert_called()
    connection_manager.connection_server.assert_called_with('peername', protocol._host, transport)
    assert isinstance(protocol._timeout_monitor, StreamAlivenessMonitor)

@patch('xml.sax.make_parser', return_value=MagicMock())
def test_connection_made_without_transport(mock_make_parser, setup_protocol):
    protocol, _ = setup_protocol
    transport = None

    with patch('pyjabber.network.server.outcoming.XMLServerOutcomingProtocol.logger') as mock_logger:
        protocol.connection_made(transport)
        mock_logger.error.assert_called_with("Invalid transport")

def test_eof_received(setup_protocol):
    protocol, connection_manager = setup_protocol
    transport = MagicMock()
    transport.get_extra_info.return_value = 'peername'
    protocol._transport = transport
    protocol._xml_parser = MagicMock()

    protocol.eof_received()

    connection_manager.disconnection_server.assert_called_with('peername')
    assert protocol._transport is None
    assert protocol._xml_parser is None


@pytest.mark.asyncio
async def test_enable_tls(setup_protocol):
    protocol, _ = setup_protocol
    mock_transport = MagicMock()
    protocol._transport = mock_transport
    mock_loop = MagicMock(spec=asyncio.AbstractEventLoop)
    mock_new_transport = MagicMock()
    mock_loop.start_tls.return_value = mock_new_transport

    # Simular connection_made para inicializar _xml_parser
    with patch('xml.sax.make_parser', return_value=MagicMock()) as mock_make_parser:
        protocol.connection_made(mock_transport)

    mock_ssl_context = MagicMock()
    with patch('ssl.create_default_context', return_value=mock_ssl_context) as mock_create_context:
        await protocol.enable_tls(mock_loop)

        mock_loop.start_tls.assert_called_once()
        assert protocol._transport == mock_new_transport

        # Verificar que OP_NO_TLSv1_3 no se ha añadido a las opciones
        mock_ssl_context.options.__ior__.assert_not_called()


@pytest.mark.asyncio
async def test_enable_tls_no_tls1_3(setup_protocol):
    protocol, _ = setup_protocol
    protocol._enable_tls1_3 = False
    mock_transport = MagicMock()
    protocol._transport = mock_transport
    mock_loop = MagicMock(spec=asyncio.AbstractEventLoop)
    mock_new_transport = MagicMock()
    mock_loop.start_tls.return_value = mock_new_transport

    # Simular connection_made para inicializar _xml_parser
    with patch('xml.sax.make_parser', return_value=MagicMock()) as mock_make_parser:
        protocol.connection_made(mock_transport)

    mock_ssl_context = MagicMock()
    mock_ssl_context.options = 0  # Inicializar options como un entero

    with patch('ssl.create_default_context', return_value=mock_ssl_context) as mock_create_context:
        await protocol.enable_tls(mock_loop)

        mock_loop.start_tls.assert_called_once()
        assert protocol._transport == mock_new_transport

        # Verificar que OP_NO_TLSv1_3 se ha añadido a las opciones
        assert mock_ssl_context.options & ssl.OP_NO_TLSv1_3 == ssl.OP_NO_TLSv1_3
if __name__ == "__main__":
    pytest.main()



from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber.utils import Singleton
from pyjabber.metadata import host


class Ping(metaclass=Singleton):
    def feed(self, jid: str, element: ET.Element):
        return ET.tostring(
            ET.Element(
                "iq",
                attrib={
                    "from": host.get(),
                    "id": element.attrib.get('id') or str(uuid4()),
                    "to": element.attrib.get('to'),
                    "type": "result",
                },
            )
        )



import pytest
from unittest.mock import Mock, MagicMock
from xml.etree import ElementTree as ET
import base64

# Ensure this path is correct
from pyjabber.stream.server.incoming.StreamServerIncomingHandler import StreamServerIncomingHandler, Signal, Stage
from pyjabber.features.StartTLSFeature import StartTLSFeature

@pytest.fixture
def mock_buffer():
    mock = Mock()
    return mock

@pytest.fixture
def mock_connection_manager():
    mock = Mock()
    return mock

@pytest.fixture
def stream_handler(mock_buffer, mock_connection_manager):
    return StreamServerIncomingHandler('mock_host', mock_buffer, 'mock_starttls', mock_connection_manager)

def test_initialization(stream_handler, mock_buffer, mock_connection_manager):
    assert stream_handler._host == 'mock_host'
    assert stream_handler._buffer == mock_buffer
    assert stream_handler._starttls == 'mock_starttls'
    assert stream_handler._connection_manager == mock_connection_manager

def test_handle_open_stream_connected(stream_handler, mock_buffer):
    stream_handler._stage = Stage.CONNECTED
    stream_handler.handle_open_stream()
    mock_buffer.write.assert_called_once()
    assert stream_handler._stage == Stage.OPENED

def test_handle_open_stream_ssl(stream_handler, mock_buffer):
    stream_handler._stage = Stage.SSL
    stream_handler.handle_open_stream()
    mock_buffer.write.assert_called_once()
    assert stream_handler._stage == Stage.SASL

def test_handle_open_stream_auth(stream_handler, mock_buffer):
    stream_handler._stage = Stage.AUTH
    result = stream_handler.handle_open_stream()
    mock_buffer.write.assert_called_once_with(b"<features xmlns='http://etherx.jabber.org/streams'/>")
    assert stream_handler._stage == Stage.READY
    assert result == Signal.DONE

def test_handle_open_stream_exception(stream_handler):
    stream_handler._stage = None
    with pytest.raises(Exception):
        stream_handler.handle_open_stream()

def test_handle_open_stream_starttls(stream_handler, mock_buffer):
    stream_handler._stage = Stage.OPENED
    stream_handler._starttls = Mock()
    starttls_element = ET.Element(StreamServerIncomingHandler.STARTTLS)
    result = stream_handler.handle_open_stream(starttls_element)
    mock_buffer.write.assert_called_once_with(StartTLSFeature().proceed_response())
    stream_handler._starttls.assert_called_once()  # Ensure _starttls was called
    assert stream_handler._stage == Stage.SSL
    assert result == Signal.RESET


def test_handle_open_stream_sasl_auth_with_mechanism(stream_handler, mock_buffer):
    stream_handler._stage = Stage.SASL
    auth_element = ET.Element(StreamServerIncomingHandler.AUTH, attrib={'mechanism': 'EXTERNAL'})
    auth_element.text = base64.b64encode(b'mock_host').decode()
    result = stream_handler.handle_open_stream(auth_element)
    mock_buffer.write.assert_called_once_with(b"<success xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>")
    assert stream_handler._stage == Stage.AUTH
    assert result == Signal.RESET

def test_handle_open_stream_sasl_auth_no_mechanism(stream_handler, mock_buffer):
    stream_handler._stage = Stage.SASL
    auth_element = ET.Element(StreamServerIncomingHandler.AUTH, attrib={'mechanism': 'EXTERNAL'})
    auth_element.text = None  # Explicitly set text to None to trigger the exception
    with pytest.raises(Exception):
        stream_handler.handle_open_stream(auth_element)


def test_handle_open_stream_sasl_auth_invalid_text(stream_handler, mock_buffer):
    stream_handler._stage = Stage.SASL
    auth_element = ET.Element(StreamServerIncomingHandler.AUTH, attrib={'mechanism': 'EXTERNAL'})
    auth_element.text = 'invalid'
    with pytest.raises(Exception):
        stream_handler.handle_open_stream(auth_element)



import pytest
from unittest.mock import Mock
from xml.etree import ElementTree as ET

# Ensure this path is correct
from pyjabber.stream.server.outcoming.StreamServerOutcomingHandler import StreamServerOutcomingHandler, Signal, Stage

@pytest.fixture
def mock_buffer():
    mock = Mock()
    return mock


@pytest.fixture
def mock_connection_manager():
    mock = Mock()
    return mock


@pytest.fixture
def stream_handler(mock_buffer, mock_connection_manager):
    return StreamServerOutcomingHandler('mock_host', mock_buffer, 'mock_starttls', mock_connection_manager)


def test_initialization(stream_handler, mock_buffer, mock_connection_manager):
    assert stream_handler._host == 'mock_host'
    assert stream_handler._buffer == mock_buffer
    assert stream_handler._starttls == 'mock_starttls'
    assert stream_handler._connection_manager == mock_connection_manager


def test_handle_open_stream_ready(stream_handler, mock_buffer, mock_connection_manager):
    stream_handler._stage = Stage.READY
    peer = ('127.0.0.1', 12345)
    mock_buffer.get_extra_info.return_value = peer

    result = stream_handler.handle_open_stream()

    mock_connection_manager.set_server_transport.assert_called_once_with(peer, mock_buffer)
    assert result == Signal.DONE


def test_handle_open_stream_features_starttls(stream_handler, mock_buffer):
    stream_handler._stage = Stage.CONNECTED
    features_element = ET.Element(stream_handler.FEATURES)
    starttls_element = ET.SubElement(features_element, stream_handler.STARTTLS)

    result = stream_handler.handle_open_stream(features_element)

    mock_buffer.write.assert_called_once_with(b"<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>")
    assert stream_handler._stage == Stage.OPENED
    assert result is None


def test_handle_open_stream_features_mechanisms(stream_handler, mock_buffer):
    mechanisms_element = ET.Element(stream_handler.MECHANISMS)
    mech_element = ET.SubElement(mechanisms_element, 'mechanism')
    mech_element.text = 'EXTERNAL'
    features_element = ET.Element(stream_handler.FEATURES)
    features_element.append(mechanisms_element)

    result = stream_handler.handle_open_stream(features_element)

    mock_buffer.write.assert_called_once_with(
        b"<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='EXTERNAL'>=</auth>")
    assert result is None


def test_handle_open_stream_proceed(stream_handler, mock_buffer):
    stream_handler._stage = Stage.OPENED
    stream_handler._starttls = Mock()  # Mock _starttls as a callable
    proceed_element = ET.Element(stream_handler.PROCEED)

    result = stream_handler.handle_open_stream(proceed_element)

    stream_handler._starttls.assert_called_once()
    assert stream_handler._stage == Stage.SSL
    assert result == Signal.RESET


def test_handle_open_stream_success(stream_handler):
    success_element = ET.Element(stream_handler.SUCCESS)

    result = stream_handler.handle_open_stream(success_element)

    assert stream_handler._stage == Stage.READY
    assert result == Signal.RESET



import os
import xml.etree.ElementTree as ET
import pytest
from unittest.mock import Mock, MagicMock

from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.server.incoming.StanzaServerIncomingHandler import StanzaServerIncomingHandler
@pytest.fixture
def mock_connection_manager():
    mock = Mock()
    mock.get_server_host.return_value = 'mock_host'
    mock.get_buffer.return_value = [Mock()]
    return mock


@pytest.fixture
def mock_buffer():
    mock = Mock()
    mock.get_extra_info.return_value = ('127.0.0.1', 12345)
    return mock


@pytest.fixture
def stanza_handler(mock_buffer, mock_connection_manager):
    return StanzaServerIncomingHandler(mock_buffer, mock_connection_manager)


def test_initialization(stanza_handler, mock_buffer, mock_connection_manager):
    assert stanza_handler._buffer == mock_buffer
    assert stanza_handler._connection_manager == mock_connection_manager
    assert stanza_handler._peername == ('127.0.0.1', 12345)
    assert stanza_handler._host == 'mock_host'
    assert isinstance(stanza_handler._presenceManager, Presence)


def test_feed_handle_iq(stanza_handler):
    iq_element = ET.Element('{jabber:client}iq')
    stanza_handler.handle_iq = Mock()

    # Redefine the functions dictionary to ensure mock is used
    stanza_handler._functions = {
        "{jabber:client}iq": stanza_handler.handle_iq,
        "{jabber:client}message": stanza_handler.handle_msg,
        "{jabber:client}presence": stanza_handler.handle_pre
    }

    stanza_handler.feed(iq_element)
    stanza_handler.handle_iq.assert_called_once_with(iq_element)


def test_feed_handle_msg(stanza_handler, mock_connection_manager):
    msg_element = ET.Element('{jabber:client}message', attrib={'to': 'user@domain'})

    mock_buffer_1 = MagicMock()
    mock_buffer_2 = MagicMock()

    mock_connection_manager.get_buffer.return_value = [[mock_buffer_1], [mock_buffer_2]]

    stanza_handler._functions = {
        "{jabber:client}iq": stanza_handler.handle_iq,
        "{jabber:client}message": stanza_handler.handle_msg,
        "{jabber:client}presence": stanza_handler.handle_pre
    }

    stanza_handler.feed(msg_element)

    mock_buffer_1.write.assert_called_once_with(ET.tostring(msg_element))
    mock_buffer_2.write.assert_called_once_with(ET.tostring(msg_element))


def test_feed_handle_pre(stanza_handler):
    pre_element = ET.Element('{jabber:client}presence')
    stanza_handler.handle_pre = Mock()

    stanza_handler._functions = {
        "{jabber:client}iq": stanza_handler.handle_iq,
        "{jabber:client}message": stanza_handler.handle_msg,
        "{jabber:client}presence": stanza_handler.handle_pre
    }

    stanza_handler.feed(pre_element)
    stanza_handler.handle_pre.assert_called_once_with(pre_element)


def test_feed_invalid_element(stanza_handler, mock_buffer):
    invalid_element = ET.Element('{jabber:client}invalid')

    stanza_handler._functions = {
        "{jabber:client}iq": stanza_handler.handle_iq,
        "{jabber:client}message": stanza_handler.handle_msg,
        "{jabber:client}presence": stanza_handler.handle_pre
    }

    stanza_handler.feed(invalid_element)
    mock_buffer.write.assert_called_once_with(SE.bad_request())


def test_handle_msg(stanza_handler, mock_connection_manager):
    msg_element = ET.Element('{jabber:client}message', attrib={'to': 'user@domain'})
    mock_receiver_buffer = [Mock()]
    mock_connection_manager.get_buffer.return_value = mock_receiver_buffer

    stanza_handler.handle_msg(msg_element)

    for buffer in mock_receiver_buffer:
        buffer[-1].write.assert_called_once_with(ET.tostring(msg_element))


def test_handle_msg(stanza_handler, mock_connection_manager):
    msg_element = ET.Element('{jabber:client}message', attrib={'to': 'user@domain'})

    mock_buffer = MagicMock()

    mock_connection_manager.get_buffer.return_value = [[mock_buffer]]

    stanza_handler.handle_msg(msg_element)

    mock_buffer.write.assert_called_once_with(ET.tostring(msg_element))

def test_handle_iq(stanza_handler):
    iq_element = ET.Element('{jabber:client}iq')
    assert stanza_handler.handle_iq(iq_element) is None



def test_handle_pre(stanza_handler):
    pre_element = ET.Element('{jabber:client}presence')
    assert stanza_handler.handle_pre(pre_element) is None




import pytest
from unittest.mock import MagicMock, patch
from xml.etree.ElementTree import Element
from pyjabber.network.server.outcoming.XMLServerOutcomingParser import XMLServerOutcomingParser
from pyjabber.stream.StanzaHandler import StanzaHandler
from pyjabber.stream.StreamHandler import Signal
from pyjabber.stream.server.outcoming.StreamServerOutcomingHandler import StreamServerOutcomingHandler
from pyjabber.utils import ClarkNotation as CN


@pytest.fixture
def setup_parser():
    buffer = MagicMock()
    starttls = MagicMock()
    connection_manager = MagicMock()
    queue_message = MagicMock()
    host = 'example.com'
    public_host = 'public.example.com'
    parser = XMLServerOutcomingParser(buffer, starttls, connection_manager, queue_message, host, public_host)
    return parser, buffer, connection_manager


def test_initialization(setup_parser):
    parser, buffer, connection_manager = setup_parser
    assert parser._host == 'example.com'
    assert parser._public_host == 'public.example.com'
    assert isinstance(parser._streamHandler, StreamServerOutcomingHandler)


@patch('pyjabber.network.server.outcoming.XMLServerOutcomingParser.Stream')
def test_initial_stream(mock_stream, setup_parser):
    parser, buffer, _ = setup_parser
    mock_stream.Stream.return_value.open_tag.return_value = b'<stream>'

    parser.initial_stream()

    mock_stream.Stream.assert_called_with(
        from_='public.example.com',
        to='example.com',
        xmlns=mock_stream.Namespaces.SERVER.value
    )
    buffer.write.assert_called_with(b'<stream>')


def test_start_element_ns(setup_parser):
    parser, _, _ = setup_parser
    name = ("http://etherx.jabber.org/streams", "stream")
    attrs = {("http://www.w3.org/XML/1998/namespace", "lang"): "en"}

    parser.startElementNS(name, None, attrs)

    assert len(parser._stack) == 1
    assert parser._stack[0].tag == '{http://etherx.jabber.org/streams}stream'


def test_start_element_ns_duplicate_stream(setup_parser):
    parser, _, _ = setup_parser
    parser._stack.append(Element('{http://etherx.jabber.org/streams}stream'))
    name = ("http://etherx.jabber.org/streams", "stream")

    with pytest.raises(Exception):
        parser.startElementNS(name, None, {})


def test_end_element_ns_stream_end(setup_parser):
    parser, buffer, _ = setup_parser
    name = "</stream:stream>"

    parser.endElementNS(name, None)

    buffer.write.assert_called_with(b'</stream>')
    assert not parser._stack


def test_end_element_ns_invalid_stanza(setup_parser):
    parser, _, _ = setup_parser
    parser._stack.append(Element('wrong_tag'))
    name = ("namespace", "element")

    with pytest.raises(Exception):
        parser.endElementNS(name, None)


@patch('pyjabber.stream.server.outcoming.StreamServerOutcomingHandler.StreamServerOutcomingHandler.handle_open_stream')
def test_end_element_ns_reset_signal(mock_handle_open_stream, setup_parser):
    parser, _, _ = setup_parser
    parser._stack.append(Element('{http://etherx.jabber.org/streams}stream'))
    parser._state = parser.StreamState.CONNECTED
    elem = Element(CN.clarkFromTuple(("namespace", "element")))
    parser._stack.append(elem)

    mock_handle_open_stream.return_value = Signal.RESET

    with patch.object(parser, 'initial_stream') as mock_initial_stream:
        parser.endElementNS(("namespace", "element"), None)
        assert not parser._stack
        mock_initial_stream.assert_called_once()


if __name__ == "__main__":
    pytest.main()



import asyncio
import os
import signal
import socket

from contextlib import closing
from loguru import logger

from pyjabber.db.database import connection
from pyjabber.network.XMLProtocol import XMLProtocol
from pyjabber.network.server.incoming.XMLServerIncomingProtocol import XMLServerIncomingProtocol
from pyjabber.network.server.outcoming.XMLServerOutcomingProtocol import XMLServerOutcomingProtocol
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.stream.QueueMessage import QueueMessage
from pyjabber.webpage.adminPage import admin_instance
from pyjabber.network import CertGenerator
from pyjabber.metadata import host as metadata_host, config_path as metadata_config_path
from pyjabber.metadata import database_path as metadata_database_path, root_path as metadata_root_path

SERVER_FILE_PATH = os.path.dirname(os.path.abspath(__file__))


class Server:
    """
        Server class

        :param host: Host for the clients connections (localhost by default)
        :param client_port: Port for client connections (5222 by default)
        :param server_port: Port for server-to-server connections (5269 by default)
        :param family: Type of AddressFamily (IPv4 or IPv6)
        :param connection_timeout: Max time without any response from a client. After that, the server will terminate the connection
        :param database_path: Path to the sqlite3 database file. If it does not exist, a new file/database will be created (site-packages/pyjabber/db by default)
        :param database_purge: Flag to indicate the reset process of the database. CAUTION! ALL THE INFORMATION WILL BE LOST
        :param enable_tls1_3: Boolean. Enables the use of TLSv1.3 in the STARTTLS process
        :param cert_path: Path to custom domain certs. By default, the server generates its own certificates for hostname
    """

    def __init__(
        self,
        host='localhost',
        client_port=5222,
        server_port=5269,
        server_out_port=5269,
        family=socket.AF_INET,
        connection_timeout=60,
        database_path=os.path.join(SERVER_FILE_PATH, 'db', 'server.db'),
        database_purge=False,
        enable_tls1_3=False,
        cert_path=None
    ):
        # Server
        self._host = host
        self._client_port = client_port
        self._server_port = server_port
        self._server_out_port = server_out_port
        self._family = family
        self._client_listener = None
        self._server_listener = None
        self._adminServer = None
        self._public_ip = None
        self._connection_timeout = connection_timeout
        self._database_path = database_path
        self._sql_init_script = os.path.join(SERVER_FILE_PATH, 'db', 'schema.sql')
        self._sql_delete_script = os.path.join(SERVER_FILE_PATH, 'db', 'delete.sql')
        self._database_purge = database_purge
        self._cert_path = cert_path

        # Client handler
        self._enable_tls1_3 = enable_tls1_3

        # Singletons
        self._connection_manager = ConnectionManager(self.task_s2s)
        self._queue_message = QueueMessage(self._connection_manager)

        metadata_host.set(host)
        metadata_database_path.set(self._database_path)
        metadata_config_path.set(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config/config.yaml'))
        metadata_root_path.set(SERVER_FILE_PATH)

    async def run_server(self):
        logger.info("Starting server...")

        if os.path.isfile(self._database_path) is False:
            logger.info("No database found. Initializing one...")
            if self._database_purge:
                logger.info("Ignoring purge database flag. No DB to purge")
            with closing(connection()) as con:
                with open(self._sql_init_script, "r") as script:
                    con.cursor().executescript(script.read())
                con.commit()
        else:
            if self._database_purge:
                logger.info("Resetting the database to default state...")
                with closing(connection()) as con:
                    with open(self._sql_delete_script, "r") as script:
                        con.cursor().executescript(script.read())
                    con.commit()

        if CertGenerator.check_hostname_cert_exists(self._host) is False and self._cert_path is None:
            CertGenerator.generate_hostname_cert(self._host)

        loop = asyncio.get_running_loop()

        lan_ip = self.query_local_ip()

        self._client_listener = await loop.create_server(
            lambda: XMLProtocol(
                namespace='jabber:client',
                host=self._host,
                connection_timeout=self._connection_timeout,
                cert_path=self._cert_path,
                enable_tls1_3=self._enable_tls1_3,
            ),
            host=[self._host, lan_ip] if lan_ip else [self._host],
            port=self._client_port,
            family=self._family
        )

        logger.info(f"Client domain => {self._host}")
        logger.info(f"Server is listening clients on {[s.getsockname() for s in self._client_listener.sockets]}")

        self._server_listener = await loop.create_server(
            lambda: XMLServerIncomingProtocol(
                namespace='jabber:server',
                host=self._host,
                connection_timeout=self._connection_timeout,
                cert_path=self._cert_path,
                enable_tls1_3=self._enable_tls1_3,
            ),
            host=["0.0.0.0"],
            port=self._server_port,
            family=self._family
        )

        logger.info(f"Server is listening servers on {[s.getsockname() for s in self._server_listener.sockets]}")
        logger.info("Server started...")

    async def stop(self):
        logger.info("Stopping server...")

        if self._client_listener and self._client_listener.is_serving():
            self._client_listener.close()
            await self._client_listener.wait_closed()

        if self._server_listener and self._server_listener.is_serving():
            self._server_listener.close()
            await self._server_listener.wait_closed()

        logger.info("Server stopped...")

    def task_s2s(self, host):
        host = host.split("@")[-1]
        loop = asyncio.get_running_loop()

        asyncio.ensure_future(self.server_connection(host, loop), loop=loop)

    async def server_connection(self, remote_host, loop):
        """
            Task to connect to another XMPP server
            :param remote_host: Host of the remote server to connect
            :param loop: Asyncio running loop. Necessary to perform task
        """
        await loop.create_connection(
            lambda: XMLServerOutcomingProtocol(
                namespace="jabber:server",
                host=remote_host,
                public_host=self._public_ip,
                connection_timeout=self._connection_timeout,
                enable_tls1_3=self._enable_tls1_3,
            ),
            host=remote_host,
            port=self._server_out_port
        )

    def query_local_ip(self):
        """
            Return the local IP of the host machine
        """
        mock_connection = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            mock_connection.connect(("8.8.8.8", 80))
            local_ip_address = mock_connection.getsockname()[0]
            return local_ip_address
        except OSError as e:
            logger.error(f"Unable to retrieve LAN IP: {e}")
        finally:
            mock_connection.close()

    def raise_exit(self):
        raise SystemExit(1)

    def start(self, debug: bool = False):
        """
            Start the already created and configuration server
            :param debug: Boolean. Enables debug mode in asyncio
        """
        loop = asyncio.get_event_loop()
        loop.set_debug(debug)

        try:
            loop.add_signal_handler(signal.SIGINT, self.raise_exit)
            loop.add_signal_handler(signal.SIGABRT, self.raise_exit)
            loop.add_signal_handler(signal.SIGTERM, self.raise_exit)
        except NotImplementedError:  # pragma: no cover
            pass

        try:
            # XMPP Server
            main_server = loop.create_task(self.run_server())
            loop.run_until_complete(main_server)

            # Control Panel Webpage | localhost:9090
            admin_server = admin_instance()
            loop.run_until_complete(admin_server)

        except (SystemExit, KeyboardInterrupt):  # pragma: no cover
            pass

        finally:
            # Cancel pending tasks
            tasks = asyncio.all_tasks(loop)
            for task in tasks:
                task.cancel()
            loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))

            # Close the server
            close_task = loop.create_task(self.stop())
            loop.run_until_complete(close_task)
            loop.close()
            asyncio.set_event_loop(None)
